###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.datatypes.date import Date
from onyx.core.database.objdb import ObjNotFound
from onyx.core.database.objdb_api import AddObj, GetObj, UpdateObj, DelObj
from onyx.core.database.objdb_api import ExistsInDatabase
from onyx.core.database.test import ufo_testcls
from onyx.core.utils.unittest import ObjDbTestCase

import unittest


# --- test API with the postgres client on TestDb
class Test_Postgres(ObjDbTestCase):
    def setUp(self):
        super().setUp()
        self.obj = ufo_testcls.ufocls(Name="test")

    def test_add_get_update_delete(self):
        self.assertRaises(ObjNotFound, GetObj, self.obj.Name, False)
        self.assertEqual(AddObj(self.obj), self.obj)

        self.assertEqual(GetObj(self.obj.Name, True), self.obj)
        self.assertEqual(GetObj(self.obj.Name, True).Birthday,
                         self.obj.Birthday)
        self.assertEqual(GetObj(self.obj.Name, True).OtherDates,
                         self.obj.OtherDates)
        self.assertEqual(GetObj(self.obj.Name, True).SimpleCurve,
                         self.obj.SimpleCurve)

        self.obj.Birthday = Date.now()
        self.assertEqual(UpdateObj(self.obj), self.obj)

        self.assertEqual(GetObj(self.obj.Name, True), self.obj)

        DelObj(self.obj)
        self.assertRaises(ObjNotFound, GetObj, self.obj.Name)

    def test_exists(self):
        self.assertFalse(ExistsInDatabase(self.obj.Name))
        AddObj(self.obj)
        self.assertTrue(ExistsInDatabase(self.obj.Name))
        DelObj(self.obj)
        self.assertFalse(ExistsInDatabase(self.obj.Name))


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..date import Date
from ..rdate import RDate

import datetime
import unittest


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- perform set-up actions, if any
        pass

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_apply_rule(self):
        zero_shift = RDate("+0d")
        for d in (datetime.date.today(), datetime.datetime.now()):
            self.assertEqual(d + zero_shift, d)

    def test_exceptions(self):
        zero_shift = RDate("+0d")
        with self.assertRaises(ValueError):
            "abc" + zero_shift

    def test_daterule(self):
        d0 = Date(2008, 1, 1)
        self.assertEqual(d0 + RDate("+e+1m"), Date(2008, 2, 29))
        self.assertEqual(d0 + RDate("+e+25d+1y+1m+0J"), Date(2009, 3, 1))
        d0 = Date(2008, 1, 5)
        self.assertEqual(d0 + RDate("+0b"), Date(2008, 1, 7))
        self.assertEqual(d0 + RDate("+5b"), Date(2008, 1, 11))
        self.assertEqual(d0 + RDate("-0b"), Date(2008, 1, 4))
        self.assertEqual(d0 + RDate("-5b"), Date(2007, 12, 31))
        d0 = Date(2008, 4, 28)
        self.assertEqual(d0 + RDate("A"), Date(2008, 1, 1))
        self.assertEqual(d0 + RDate("q"), Date(2008, 4, 1))
        self.assertEqual(d0 + RDate("Q"), Date(2008, 6, 30))
        self.assertEqual(d0 + RDate("E"), Date(2008, 12, 31))
        d0 = Date(2008, 6, 18)
        self.assertEqual(d0 + RDate("-1M"), Date(2008, 6, 16))
        self.assertEqual(d0 + RDate("+1M"), Date(2008, 6, 23))
        self.assertEqual(d0 + RDate("-1F"), Date(2008, 6, 13))
        self.assertEqual(d0 + RDate("+1F"), Date(2008, 6, 20))
        # --- last monday and friday of the month
        self.assertEqual(d0 + RDate("+e-1M"), Date(2008, 6, 30))
        self.assertEqual(d0 + RDate("+e-1F"), Date(2008, 6, 27))

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import IntField

from onyx.core.depgraph.graph_api import ValueType, CreateInMemory
from onyx.core.depgraph.graph_api import GetVal, GetNode, InvalidateNode
from onyx.core.utils.unittest import OnyxTestCase

import unittest


###############################################################################
class Fibonacci(UfoBase):
    fib0 = IntField(default=0)
    fib1 = IntField(default=1)

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def fib(self, graph, k):
        if k == 0:
            return graph(self, "fib0")
        elif k == 1:
            return graph(self, "fib1")
        else:
            return graph(self, "fib", k-1) + graph(self, "fib", k-2)


###############################################################################
class TestFibonacci(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = CreateInMemory(Fibonacci(Name="fib"))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def tearDown(self):
        super().tearDown()

    # -------------------------------------------------------------------------
    def test_initial_values(self):
        self.assertEqual(GetVal(self.name, "fib0"), 0)
        self.assertEqual(GetVal(self.name, "fib1"), 1)

    # -------------------------------------------------------------------------
    def test_seq(self):
        self.assertEqual(GetVal(self.name, "fib", 0), 0)
        self.assertEqual(GetVal(self.name, "fib", 1), 1)
        self.assertEqual(GetVal(self.name, "fib", 2), 1)
        self.assertEqual(GetVal(self.name, "fib", 3), 2)
        self.assertEqual(GetVal(self.name, "fib", 4), 3)
        self.assertEqual(GetVal(self.name, "fib", 5), 5)
        self.assertEqual(GetVal(self.name, "fib", 6), 8)
        self.assertEqual(GetVal(self.name, "fib", 7), 13)
        self.assertEqual(GetVal(self.name, "fib", 8), 21)
        self.assertEqual(GetVal(self.name, "fib", 9), 34)

    # -------------------------------------------------------------------------
    def test_graph(self):
        """
        Calculate a large fibonacci number and then perform tests on the
        generated graph.
        """
        self.assertEqual(GetVal(self.name, "fib", 10), 55)

        parents = {('fib', 'fib', (9,)), ('fib', 'fib', (10,))}
        children = {('fib', 'fib', (6,)), ('fib', 'fib', (7,))}
        node_8 = GetNode(self.name, "fib", (8,))

        self.assertTrue(node_8.valid)
        self.assertEqual(node_8.value, 21)
        self.assertEqual(node_8.parents, parents)
        self.assertEqual(node_8.children, children)

        # --- invalidate node 8 and check again the state of the graph
        InvalidateNode(self.name, "fib", (8,))

        # --- invalidation sets the node to invalid, but doesn't afftect the
        #     topology of the graph. Such changes take place just before
        #     recalculating the value of an invalid node (the graph is
        #     extremely lazy).
        self.assertFalse(node_8.valid)
        self.assertEqual(node_8.value, 21)
        self.assertEqual(node_8.parents, parents)
        self.assertEqual(node_8.children, children)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..datatypes.date import Date
from ..datatypes.gcurve import GCurve
from ..datatypes.curve import Curve
from ..database.ufo_base import UfoBase, BaseField, custom_decoder
from ..database.ufo_fields import FloatField
from ..database.objdb import ObjNotFound
from ..database.objdb_api import ObjDbQuery
from .graph_api import ValueType, GetVal

from .. import database as onyx_db

import json

__all__ = [
    "Archivable",
    "MktIndirectionFactory",
    "EnforceArchivableEntitlements",
]

GET_HISTORY_QUERY = """
SELECT Date, Value FROM Archive
WHERE Name=%s AND Attribute=%s AND Date BETWEEN %s AND %s ORDER BY Date;"""

GET_MOST_RECENT = """
SELECT MAX(Date) FROM Archive
WHERE Name=%s AND Attribute=%s AND Date <= %s"""


###############################################################################
class Archivable(UfoBase):
    """
    Base class for all archivable UFO objects.
    """

    # -------------------------------------------------------------------------
    def last_before(self, attr, date=None):
        """
        Description:
            Return the date of the most recent archived record before a given
            date.
        Inputs:
            attr - the required attribute
            date - look for dated records with date earlier than this value.
        Returns:
            A Date.
        """
        date = date or Date.high_date()
        res = ObjDbQuery(GET_MOST_RECENT,
                         (self.Name, attr, date), attr="fetchone")
        if res.max is None:
            raise ObjNotFound("{0!s} not found for any date earlier"
                              "than {1:s}".format((self.Name, attr), date))
        else:
            return res.max

    # -------------------------------------------------------------------------
    def get_dated(self, attr, date, strict=False):
        """
        Description:
            Get the value of an archived attribute as of a given date.
        Inputs:
            attr   - the attribute to archive
            date   - the archive date
            strict - if true, raise an ObjNotFound exception if a record cannot
                     be found for the required date
        Returns:
            A (Date, value) tuple
        """
        # --- get the instance of the field descriptor that is storing
        #     the attribute descriptor itself
        field = getattr(self.__class__, attr).field

        # --- load from database and de-serialize value
        date, value = onyx_db.obj_clt.get_dated(self.Name,
                                                attr, date, strict=strict)

        return date, field.from_json(value)

    # -------------------------------------------------------------------------
    def set_dated(self, attr, date, value, overwrite=False):
        """
        Description:
            Set the value of an archived attribute.
        Inputs:
            attr      - the attribute to archive
            date      - the archive date
            value     - the value to archive
            overwrite - if true, existing records can be overwritten
        """
        # --- get the instance of the field descriptor that is storing
        #     the attribute descriptor itself
        field = getattr(self.__class__, attr).field
        # --- make sure value is compatible with the field_type
        field.validate(value)
        # --- serialize value
        value = field.to_json(value)

        onyx_db.obj_clt.set_dated(self.Name, attr,
                                  date, value, overwrite=overwrite)

    # -------------------------------------------------------------------------
    def delete(self):
        """
        Invoked when the archived object is deleted from database.
        """
        ObjDbQuery("DELETE FROM Archive WHERE Name=%s", parms=(self.Name,))

    # -------------------------------------------------------------------------
    def get_history(self, attr, start=None, end=None):
        start = start or Date.low_date()
        end = end or Date.high_date()

        # --- get the instance of the field descriptor that is stored
        #     the VT descriptor itself
        field = getattr(self.__class__, attr).field

        parms = (self.Name, attr, start, end)
        rows = ObjDbQuery(GET_HISTORY_QUERY, parms, attr="fetchall")

        # --- usually json.loads is called by get_dated before returning the
        #     value but here we are processing results from a raw query
        def convert(val):
            return field.from_json(json.loads(val, cls=custom_decoder))

        knots = [(Date.parse(r[0]), convert(r[1])) for r in rows]

        if isinstance(field, FloatField):
            return Curve([d for d, v in knots], [v for d, v in knots])
        else:
            return GCurve([d for d, v in knots], [v for d, v in knots])


###############################################################################
class MktIndirectionFactory(object):
    """
    Description:
        ValueType factory that implements a descriptor protocol used for market
        data indirection.
    Inputs:
        field_type - A valid field-type constructor (must be a subclass of
                     BaseField).
        obj        - name of the object exposing value types with the current
                     market data date and a flag dictating if archived values
                     need to be looked up in strict mode or not (default is
                     "Database").
        mdd_vt     - name of the value type for market data date (default
                     is "MktDataDate").
        fs_vt      - name of the value type for force strict flag (default
                     is "ForceStrict").

    THIS DECORATOR CAN ONLY BE APPLIED TO METHODS OF A SUBCLASS OF Archivable.

    Typical use is as follows:

        @MktIndirectionFactory(FloatField)
        def MarketizedVt(self, graph): pass

    MarketizedVt is created as a Property ValueType that fetches archived data
    by accessing the corresponding dated record. Consistency with the field
    type is enforced.
    """
    # -------------------------------------------------------------------------
    def __init__(self, field_type,
                 obj="Database", mdd_vt="MktDataDate", fs_vt="ForceStrict"):
        if not issubclass(field_type, BaseField):
            raise ValueError("field_type must be an instance of a "
                             "class derived from BaseField. Got {0!s} "
                             "instead".format(field_type.__class__))

        self.field = field_type()
        self.obj = obj
        self.mdd_vt = mdd_vt
        self.fs_vt = fs_vt

    # -------------------------------------------------------------------------
    def __call__(self, func):
        vt_name = func.__name__

        # --- descriptor protocol: getter
        def getter(instance, graph):
            mdd = graph(self.obj, self.mdd_vt)  # market data date
            fs = graph(self.obj, self.fs_vt)  # force strict
            _, value = instance.get_dated(vt_name, mdd, fs)
            return value

        getter.__name__ = vt_name

        # --- create a property descriptor ValueType and add the appropriate
        #     field attribute to it
        descriptor = ValueType()(getter)
        descriptor.field = self.field

        return descriptor


###############################################################################
class EnforceArchivableEntitlements(object):
    """
    Description:
        Decorator class used to enforce compliance with Archivable
        entitlements.
    Inputs:
        obj - target object as defined in database (or memory)
        VT  - target value type (a stored attribute or a method of a ufo
              class decorated by @ValueType())

    Typical use is as follows:

        @EnforceArchivableEntitlements("Database", "ArchivedOverwritable")
        class my_ufo_class(Archivable):
        ...
    """
    # -------------------------------------------------------------------------
    def __init__(self, obj, VT):
        self.obj = obj
        self.VT = VT

    # -------------------------------------------------------------------------
    def __call__(self, cls):
        if not issubclass(cls, Archivable):
            raise ValueError("enforce_archivable_entitlements can "
                             "only be applied to a subclass of Archivable")

        # --- override set_dated to enforce compliance with the required value
        #     type
        def set_dated(instance, attr, date, value):
            overwrite = GetVal(self.obj, self.VT)
            instance.__set_dated_raw(attr, date, value, overwrite)

        # --- rename original set_dated and replace it with a new version
        cls.__set_dated_raw = cls.set_dated
        cls.set_dated = set_dated

        return cls



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import IntField, FieldError
from onyx.core.database.objdb import ObjNotFound
from onyx.core.database.objdb_api import GetObj
from onyx.core.depgraph.graph import GraphError
from onyx.core.depgraph.graph_api import ValueType, GetVal, SetVal
from onyx.core.depgraph.graph_api import CreateInMemory, InvalidateNode
from onyx.core.depgraph.ufo_functions import RetainedFactory

from onyx.core.utils.unittest import OnyxTestCase

import time
import unittest

ATTR1 = 333
ATTR2 = 666


###############################################################################
class timer(object):
    # -------------------------------------------------------------------------
    def __enter__(self):
        self.start = time.time()
        return self

    # -------------------------------------------------------------------------
    def __exit__(self, *args):
        self.elapsed = time.time() - self.start


###############################################################################
class test_cls(UfoBase):
    attr1 = IntField(default=ATTR1)
    attr2 = IntField(default=ATTR2)

    # -------------------------------------------------------------------------
    @ValueType()
    def A(self, graph):
        return 1.0 + graph(self, "B") + graph(self, "C", 1, 2)

    # -------------------------------------------------------------------------
    @ValueType()
    def B(self, graph):
        # --- this is a slow method
        time.sleep(2.0)
        return graph(self, "attr1")

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def C(self, graph, x, y):
        # --- this is a slow calculated method
        time.sleep(2.0)
        return x + y + graph(self, "attr2")

    # -------------------------------------------------------------------------
    @RetainedFactory()
    def D(self, graph):
        return graph(self, "attr2") / graph(self, "attr1")

    # -------------------------------------------------------------------------
    @ValueType()
    def parent(self, graph):
        return graph(self, "D")*2.0


###############################################################################
class TestGraph(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = CreateInMemory(test_cls(Name="test"))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def tearDown(self):
        super().tearDown()

    # -------------------------------------------------------------------------
    def test_GetVal(self):
        # --- retrive stored attributes by instance name
        self.assertEqual(GetVal(self.name, "attr1"), ATTR1)
        self.assertEqual(GetVal(self.name, "attr2"), ATTR2)

        # --- retrive stored attributes by instance
        self.assertEqual(GetVal(self.instance, "attr1"), ATTR1)
        self.assertEqual(GetVal(self.instance, "attr2"), ATTR2)

        # --- retrive other value types
        self.assertEqual(GetVal(self.name, "A"), 1003.0)
        self.assertEqual(GetVal(self.name, "C", 0, 0), ATTR2)
        self.assertEqual(GetVal(self.name, "C", 1, 0), ATTR2 + 1)
        self.assertEqual(GetVal(self.name, "C", 0, 2), ATTR2 + 2)
        self.assertEqual(GetVal(self.name, "D"), 2.0)

    # -------------------------------------------------------------------------
    def test_SetVal(self):
        self.assertEqual(GetVal(self.name, "attr1"), ATTR1)
        SetVal(self.name, "attr1", 999)
        self.assertEqual(GetVal(self.name, "attr1"), 999)
        self.assertEqual(GetObj(self.name).attr1, 999)
        # --- set a retained ValueType. we first call the parent node so that D
        #     has parents and children.
        self.assertEqual(GetVal(self.name, "D"), ATTR2 / 999)
        self.assertEqual(GetVal(self.name, "parent"), (ATTR2 / 999) * 2.0)
        SetVal(self.name, "D", "a string")
        self.assertEqual(GetVal(self.name, "D"), "a string")

    # -------------------------------------------------------------------------
    def test_caching(self):
        # --- compute A the first time
        with timer() as t:
            val = GetVal(self.name, "A")
        self.assertEqual(val, 1003.0)
        self.assertAlmostEqual(t.elapsed, 4.0, 1)

        # --- compute A a second time: result is cached
        with timer() as t:
            val = GetVal(self.name, "A")
        self.assertEqual(val, 1003.0)
        self.assertAlmostEqual(t.elapsed, 0.0, 1)

        # --- invalidate B and recalculate A: should only have to recalculate
        #     one of the two children
        InvalidateNode(self.name, "B")
        with timer() as t:
            val = GetVal(self.name, "A")
        self.assertEqual(val, 1003.0)
        self.assertAlmostEqual(t.elapsed, 2.0, 1)

        # --- try once more: should be cached again
        with timer() as t:
            val = GetVal(self.name, "A")
        self.assertEqual(val, 1003.0)
        self.assertAlmostEqual(t.elapsed, 0.0, 1)

    # -------------------------------------------------------------------------
    def test_exceptions(self):
        # --- object doesn't exist
        self.assertRaises(ObjNotFound, GetVal, "xxx", "xxx")
        # --- object exists, but not the ValueType
        self.assertRaises(AttributeError, GetVal, self.name, "xxx")
        # --- VT is settable but wrong type
        self.assertRaises(FieldError, SetVal, self.name, "attr1", None)
        # --- VT not settable
        self.assertRaises(GraphError, SetVal, self.name, "A", None)
        self.assertRaises(GraphError, SetVal, self.name, "B", None)
        self.assertRaises(GraphError, SetVal, self.name, "C", None)
        # --- callable node invalidated incorrectly
        self.assertRaises(GraphError, InvalidateNode, self.name, "C")


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ...datatypes.date import Date
from ...datatypes.curve import Curve
from ..curve_fns import Interpolate, ApplyAdjustments

import numpy as np
import unittest


# --- unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- create a first curve by passing arrays of dates and values
        self.dates = [Date.parse(d)
                      for d in ["F08", "G08", "H08", "J08", "K08", "M08",
                                "N08", "Q08", "U08", "V08", "X08", "Z08"]]
        self.values = np.cumsum(np.ones(12))
        self.refcrv = Curve(self.dates, self.values)

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_interpolate(self):
        dt = Date(2009, 1, 1)
        self.assertEqual(12, Interpolate(self.refcrv, dt))
        self.assertEqual(12, Interpolate(self.refcrv, dt, "Step"))
        self.assertEqual(12, Interpolate(self.refcrv, dt, "Linear"))

        dt = Date(2007, 1, 1)
        self.assertEqual(1, Interpolate(self.refcrv, dt))
        self.assertEqual(1, Interpolate(self.refcrv, dt, "Step"))
        self.assertEqual(1, Interpolate(self.refcrv, dt, "Linear"))

        dt = Date(2008, 11, 16)
        self.assertEqual(11, Interpolate(self.refcrv, dt))
        self.assertEqual(11, Interpolate(self.refcrv, dt, "Step"))
        self.assertEqual(11.5, Interpolate(self.refcrv, dt, "Linear"))

    def test_apply_adjustments(self):
        adj_crv = Curve([Date(2008, 6, 15)], [0.1])
        ref_crv = Curve(self.refcrv.dates,
                        [0.1, 0.2, 0.3, 0.4, 0.5, 0.6,
                         7.0, 8.0, 9.0, 10.0, 11.0, 12.0])

        adjusted = ApplyAdjustments(self.refcrv, adj_crv)

        self.assertTrue(np.all(adjusted.dates == ref_crv.dates))
        self.assertTrue(np.allclose(adjusted.values, ref_crv.values))

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import datetime
import dateutil.parser

__all__ = ["Date", "DateError"]

# --- constants used for conversion from/to numerical value
HOURS_PER_DAY = 24.0
MINUTES_PER_DAY = 60.0*HOURS_PER_DAY
SECONDS_PER_DAY = 60.0*MINUTES_PER_DAY
MUSECONDS_PER_DAY = 1e6*SECONDS_PER_DAY

# --- letter to month conversion
L2M = {
    "F":  1,
    "G":  2,
    "H":  3,
    "J":  4,
    "K":  5,
    "M":  6,
    "N":  7,
    "Q":  8,
    "U":  9,
    "V": 10,
    "X": 11,
    "Z": 12,
}

# --- constant used to distinguish date from datetime
DT0 = datetime.time(0, 0)


###############################################################################
class DateError(Exception):
    """
    Base class for all Date exceptions.
    """
    pass


###############################################################################
class Date(datetime.datetime):
    """
    Sublass of the datetime.datetime class implementing a few more methods.
    """
    # -------------------------------------------------------------------------
    def __str__(self):
        if self.time() == DT0:
            # --- d is a date
            return self.strftime("%d-%b-%Y")
        else:
            # --- d is a date-time
            return self.strftime("%d-%b-%Y %H:%M:%S")

    # -------------------------------------------------------------------------
    def __hash__(self):
        # --- FIXME: for some reason in python3.4 we need to redefine...
        return super().__hash__()

    # -------------------------------------------------------------------------
    #  methods for date comparisons. superclass methods are extended so that
    #  comparisons with datetime.date instances are possible

    def __eq__(self, other):
        if other.__class__ == datetime.date:
            return self.date().__eq__(other)
        else:
            return super().__eq__(other)

    def __ne__(self, other):
        if other.__class__ == datetime.date:
            return self.date().__ne__(other)
        else:
            return super().__ne__(other)

    def __lt__(self, other):
        if other.__class__ == datetime.date:
            return self.date().__lt__(other)
        else:
            return super().__lt__(other)

    def __le__(self, other):
        if other.__class__ == datetime.date:
            return self.date().__le__(other)
        else:
            return super().__le__(other)

    def __gt__(self, other):
        if other.__class__ == datetime.date:
            return self.date().__gt__(other)
        else:
            return super().__gt__(other)

    def __ge__(self, other):
        if other.__class__ == datetime.date:
            return self.date().__ge__(other)
        else:
            return super().__ge__(other)

    # -------------------------------------------------------------------------
    @property
    def ordinal(self):
        """
        Return datetime in float ordinal (currently ignores time zones).
        """
        base = float(self.toordinal())
        if self.time() != DT0:
            base += (self.hour / HOURS_PER_DAY +
                     self.minute / MINUTES_PER_DAY +
                     self.second / SECONDS_PER_DAY +
                     self.microsecond / MUSECONDS_PER_DAY)
        return base

    # -------------------------------------------------------------------------
    def is_weekday(self, weekday):
        """
        Test for a particular day in the week.
        """
        return self.weekday() is weekday

    # -------------------------------------------------------------------------
    def eod(self):
        """
        Return a new Date corresponding to the end of day for the same day.
        """
        return self.__class__(self.year, self.month, self.day, 23, 59, 59)

    # -------------------------------------------------------------------------
    #  lower and upper limits to valid dates

    @classmethod
    def low_date(cls):
        return cls(1970, 1, 1)

    @classmethod
    def high_date(cls):
        return cls(2069, 12, 31)

    # -------------------------------------------------------------------------
    #  Date constructors

    @classmethod
    def today(cls):
        return cls.fromordinal(datetime.date.today().toordinal())

    @classmethod
    def parse(cls, d, day_first=False):
        """
           Valid Dates can be created as follows:
            - from LYY string
            - from DDMMMYY string or any other format accepted by
                   dateutil.parser.parse()
            - from date/datetime object
        """
        if isinstance(d, str):
            if len(d) == 3 and d[0] in L2M:
                # --- this is a LYY string
                yy = int(d[1:])
                mm = L2M[d[0]]
                # --- years from 70 to 99 are expected to be in 1900
                yy = 2000 + yy if yy < 70 else 1900 + yy
                # --- construct date from month and year
                return cls(yy, mm, 1)
            elif len(d) == 7:
                # --- this is a date in the DDMMMYY format
                d = datetime.datetime.strptime(d, "%d%b%y")
                return cls(d.year, d.month, d.day)
            else:
                # --- use dateutil parser
                try:
                    d = dateutil.parser.parse(d, dayfirst=day_first)
                    if d.hour or d.minute or d.second or d.microsecond:
                        return cls(d.year, d.month, d.day,
                                   d.hour, d.minute, d.second, d.microsecond)
                    else:
                        return cls(d.year, d.month, d.day)
                except ValueError:
                    raise ValueError("Unrecognized "
                                     "string format: {0:s}".format(d))

        elif isinstance(d, datetime.datetime):
            # --- N.B.: datetime is a subclass of date, hence test for datetime
            #           first
            return cls(d.year, d.month, d.day,
                       d.hour, d.minute, d.second, d.microsecond)

        elif isinstance(d, datetime.date):
            return cls(d.year, d.month, d.day)

        elif isinstance(d, (int, float)):
            # --- NB: conversion is done up to the second: microseconds will be
            #         ignored
            int_d = int(d)
            dt = datetime.datetime.fromordinal(int_d)

            remainder = float(d) - int_d
            hour, remainder = divmod(24*remainder, 1)
            minute, remainder = divmod(60*remainder, 1)
            second, remainder = divmod(60*remainder, 1)

            return cls(dt.year, dt.month, dt.day,
                       int(hour), int(minute), int(second))

        else:
            raise ValueError("Unrecognized format: "
                             "{0!s}, {1!s}".format(d, type(d)))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .graph_api import ValueType

import types

__all__ = [
    "RetainedFactory",
    "InheritAsProperty",
    "DiscardInheritedAttribute",
]


###############################################################################
class RetainedFactory(object):
    """
    Description:
        ValueType factory that implements a descriptor protocol used to
        represent pseudo-attribute VTs, i.e. VTs that can be set but are not
        persisted in database.
    Inputs:
        default - if None use the VT function to calculate the current value,
                  otherwise use default.

    Typical use cases are:

    1) for the implementation of retained properties (such as Spot) where the
       typical syntax is as follows:

           @RetainedFactory()
           def Spot(self, graph):
               ...

    2) for in the implementation of edit screens where the typical
       syntax is as follows:

            @RetainedFactory(default=...)
            def MyVt(self, graph): pass
    """
    # -------------------------------------------------------------------------
    def __init__(self, default=None):
        self.value = default

    # -------------------------------------------------------------------------
    def __call__(self, func):
        def getter(instance, graph):
            if self.value is None:
                return func(instance, graph)
            else:
                return self.value

        def setter(instance, graph, value):
            self.value = value

        # --- return a settable descriptor ValueType
        return ValueType("Settable", getter, setter)(func)


###############################################################################
class InheritAsProperty(object):
    """
    Description:
        Class decorator used to replace one or more stored attributes of the
        super class with properties pointing to the same attribute of another
        object.
    Inputs:
        attrs      - a list of stored attributes of the super-class that are
                     to be replaced by Property value types.
        value_type - a pointer to the instance of the ufo class from which we
                     are inheriting stored attributes as properties.
                     Each stored attribute in attrs will be replaced with by
                     property VT.
    """
    template = ("def template(self, graph):\n"
                "    return graph(graph(self, '{0:s}'), '{1:s}')")

    # -------------------------------------------------------------------------
    def __init__(self, attrs, value_type):
        self.attrs = attrs
        self.value_type = value_type

    # -------------------------------------------------------------------------
    def __call__(self, cls):
        for attr in self.attrs:
            # --- discard from set of StoredAttrs attributes (discard does
            #     nothing if cls doesn't have such attribute)
            cls._json_fields.discard(attr)
            cls.StoredAttrs.discard(attr)

            # --- create Property-ValueType
            mod = types.ModuleType("__templates")
            exec(self.template.format(self.value_type, attr), mod.__dict__)
            func = types.FunctionType(mod.template.__code__, {}, name=attr)
            setattr(cls, attr, ValueType()(func))

        return cls


###############################################################################
class DiscardInheritedAttribute(object):
    """
    Description:
        Class decorator used to dis-inherit one or more stored attributes of
        the super class, ususally so that they can be re-implemented as
        properties.
    Inputs:
        attrs - a list of stored attributes of the super-class that are
                to be discarded.
    """
    # -------------------------------------------------------------------------
    def __init__(self, attrs):
        self.attrs = attrs

    # -------------------------------------------------------------------------
    def __call__(self, cls):
        for attr in self.attrs:
            # --- discard from set of StoredAttrs attributes (discard does
            #     nothing if cls doesn't have such attribute)
            cls._json_fields.discard(attr)
            cls.StoredAttrs.discard(attr)

        return cls



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..datatypes.curve import Curve
from ..datatypes.hlocv import HlocvCurve
from ..datatypes.date import Date
from .objdb import typecast_to_Date

import psycopg2 as psql
import psycopg2.extras as psql_extras
import psycopg2.extensions as psql_ext
import numpy as np
import sys

__all__ = ["TsNotFound", "TsDbError", "TsDbClient"]

# --- column names for HLOCVCurves
HLOCV_FIELDS = ("High", "Low", "Open", "Close", "Volume")

# --- define standard query strings
QUERY_TS_EXISTS = """
SELECT EXISTS (SELECT 1 FROM {0:s} WHERE Name=%s) AS "exists";"""


###############################################################################
class TsNotFound(Exception):
    pass


###############################################################################
class TsDbError(Exception):
    pass


# -----------------------------------------------------------------------------
def print_error(cursor):
    print("QUERY:\n{0:s}".format(cursor.query), file=sys.stderr)


###############################################################################
class TsDbClient(object):
    """
    Description:
        TsDb database, client class. It exposes a minimal api.
    """
    core_tables = ("Curves", "HLOCVCurves")

    # -------------------------------------------------------------------------
    def __init__(self, database, user, host=None, check=True):
        self.dbname = database
        self.user = user
        self.host = host

        self.conn = psql.connect(host=host, database=database, user=user,
                                 cursor_factory=psql_extras.NamedTupleCursor)
        self.conn.set_isolation_level(psql_ext.ISOLATION_LEVEL_AUTOCOMMIT)

        # --- register typecast to Date for date, timpestamp, timestamptz
        curs = self.conn.cursor()
        curs.execute("SELECT NULL::date, NULL::timestamp, NULL::timestamptz;")
        oids = tuple([col.type_code for col in curs.description])
        psql_ext.register_type(psql_ext.new_type(oids, "Date",
                                                 typecast_to_Date))

        # --- validation: without the following tables in the backend the
        #                 client doesn't work properly.
        if check:
            curs = self.conn.cursor()
            for table in self.core_tables:
                try:
                    curs.execute("SELECT 1 FROM {0:s} LIMIT 1;".format(table))
                except psql.ProgrammingError:
                    raise RuntimeError("Table {0:s} is missing for "
                                       "{1:s}".format(table, self.dbname))

    # -------------------------------------------------------------------------
    def initialize(self):
        curs = self.conn.cursor()
        curs.execute("""
-- Curves table
CREATE TABLE
Curves (
    Date  timestamptz      NOT NULL,
    Name  varchar(64)      NOT NULL,
    Value double precision NOT NULL,
    CONSTRAINT crv_date_name UNIQUE (Date, Name) );

CREATE INDEX crv_by_name_idx ON Curves (Name);

-- HlocvCurves table
CREATE TABLE
HlocvCurves (
    Date   timestamptz      NOT NULL,
    Name   varchar(64)      NOT NULL,
    High   double precision,
    Low    double precision,
    Open   double precision,
    Close  double precision NOT NULL,
    Volume double precision,
    CONSTRAINT hlocv_date_name UNIQUE (Date, Name) );

CREATE INDEX hlocv_by_name_idx ON HlocvCurves (Name);""")

    # -------------------------------------------------------------------------
    def cleanup(self):
        curs = self.conn.cursor()
        curs.execute("""
DROP TABLE IF EXISTS Curves CASCADE;
DROP TABLE IF EXISTS HlocvCurves CASCADE;""")

    # -------------------------------------------------------------------------
    def restart(self):
        self.close()
        self.conn = psql.connect(host=self.host,
                                 database=self.dbname, user=self.user)
        self.conn.set_isolation_level(psql_ext.ISOLATION_LEVEL_AUTOCOMMIT)

    # -------------------------------------------------------------------------
    def close(self):
        if hasattr(self, "conn") and not self.conn.closed:
            self.conn.close()

    # -------------------------------------------------------------------------
    def get_row_by(self, table, name, date=None, strict=False):
        if date is None:
            query = """SELECT * FROM {0:s} WHERE Name=%s AND Date IN
                       (SELECT MAX(Date) FROM {0:s} WHERE Name=%s);"""
            parms = (name, name)
        elif strict:
            query = "SELECT * FROM {0:s} WHERE Name=%s AND Date=%s;"
            parms = (name, date)
        else:
            query = """SELECT * FROM {0:s} WHERE Name=%s AND Date IN
                       (SELECT MAX(Date)
                        FROM {0:s} WHERE Name=%s AND Date<=%s);"""
            parms = (name, name, date)

        curs = self.conn.cursor()
        try:
            curs.execute(query.format(table), parms)
        except psql.OperationalError:
            print_error(curs)
            raise

        row = curs.fetchone()
        curs.close()

        if row is None:
            if strict and date is not None:
                raise TsNotFound("Row not found for curve {0:s} in {1:s}"
                                 "and date {2:s}".format(name, table, date))
            else:
                raise TsNotFound("No rows were found "
                                 "for {0:s} in {1:s}".format(name, table))
        else:
            return row

    # -------------------------------------------------------------------------
    def get_curve(self, name, start=None, end=None):
        if start is None:
            if end is None:
                # --- neither start nor end are specified
                parms = name,
                query = """SELECT Date, Value FROM Curves
                           WHERE Name=%s ORDER BY Date;"""
                errmsg = "No data found in Curves for {0:s}".format(name)
            else:
                # --- end is specified, start is not
                parms = name, end
                query = """SELECT Date, Value FROM Curves
                           WHERE Name=%s AND Date<=%s ORDER BY Date;"""
                errmsg = ("No data found in "
                          "Curves for {0:s}, end = {1!s}").format(name, end)
        elif end is None:
            # --- start is specified, end is not
            parms = name, start
            query = """SELECT Date, Value FROM Curves
                       WHERE Name=%s AND Date>=%s ORDER BY Date;"""
            errmsg = ("No data found in Curves "
                      "for {0:s}, start = {1!s}").format(name, start)
        else:
            # --- both start and end are specified
            parms = name, start, end
            query = """SELECT Date, Value FROM Curves
                       WHERE Name=%s AND Date BETWEEN %s AND %s
                       ORDER BY Date;"""
            errmsg = ("No data found in Curves for {0:s}, "
                      "start = {1!s} end = {2!s}").format(name, start, end)

        # --- create a cursor, check if ts exists and execute query
        curs = self.conn.cursor()
        try:
            curs.execute(QUERY_TS_EXISTS.format("Curves"), (name, ))
        except psql.OperationalError:
            print_error(curs)
            raise

        if not curs.fetchone().exists:
            raise TsNotFound("Time Series {0:s} not "
                             "found in Curves table".format(name))

        try:
            curs.execute(query, parms)
        except psql.OperationalError:
            print_error(curs)
            raise
        else:
            # --- populate a curve with the query's results
            knts = [(Date.parse(r.date), r.value) for r in curs.fetchall()]
            crv = Curve.create_raw([d for d, v in knts], [v for d, v in knts])
        finally:
            curs.close()

        # --- raise exception if curve is empty
        if not len(crv):
            raise TsNotFound(errmsg)

        return crv

    # -------------------------------------------------------------------------
    def get_hlocv(self, name, start=None, end=None, field=None):
        if start is None:
            if end is None:
                # --- neither start nor end are specified
                parms = name,
                query = """SELECT Date, High, Low, Open, Close, Volume
                           FROM HLOCVCurves
                           WHERE Name=%s ORDER BY Date;"""
                errmsg = "No data found in HLOCVCurves for {0:s}".format(name)
            else:
                # --- end is specified, start is not
                parms = name, end
                query = """SELECT  Date, High, Low, Open, Close, Volume
                           FROM HLOCVCurves
                           WHERE Name=%s AND Date<=%s ORDER BY Date;"""
                errmsg = ("No data found in HLOCVCurves "
                          "for {0:s}, end = {1!s}").format(name, end)
        elif end is None:
            # --- start is specified, end is not
            parms = name, start
            query = """SELECT Date, High, Low, Open, Close, Volume
                       FROM HLOCVCurves
                       WHERE Name=%s AND Date>=%s ORDER BY Date;"""
            errmsg = ("No data found in HLOCVCurves "
                      "for {0:s}, start = {1!s}").format(name, start)
        else:
            # --- both start and end are specified
            parms = name, start, end
            query = """SELECT Date, High, Low, Open, Close, Volume
                       FROM HLOCVCurves
                       WHERE Name=%s AND Date BETWEEN %s AND %s
                       ORDER BY Date;"""
            errmsg = ("No data found in HLOCVCurves for {0:s}, "
                      "start = {1!s} end = {2!s}").format(name, start, end)

        # --- create a cursor, check if ts exists and execute query
        curs = self.conn.cursor()
        try:
            curs.execute(QUERY_TS_EXISTS.format("HLOCVCurves"), (name, ))
        except psql.OperationalError:
            print_error(curs)
            raise
        if not curs.fetchone().exists:
            raise TsNotFound("Time Series {0:s} not found "
                             "in HLOCVCurves table".format(name))

        try:
            curs.execute(query, parms)
        except psql.OperationalError:
            print_error(curs)
            raise
        else:
            if field is None:
                knts = [(Date.parse(r.date),
                        [x or np.nan for x in r[1:]]) for r in curs.fetchall()]
                crv = HlocvCurve.create_raw([d for d, v in knts],
                                            [v for d, v in knts])

            else:
                idx = 1 + HLOCV_FIELDS.index(field)
                knts = [(Date.parse(r.date),
                         r[idx] or np.nan) for r in curs.fetchall()]
                crv = Curve.create_raw([d for d, v in knts],
                                       [v for d, v in knts])

        finally:
            curs.close()

        # --- raise exception if curve is empty
        if not len(crv):
            raise TsNotFound(errmsg)

        return crv

    # -------------------------------------------------------------------------
    def upsert_curve(self, name, crv, crv_type="CRV"):
        if crv_type == "HLOCV":
            def is_equal(v1, v2): return np.all(v1 == v2)  # analysis:ignore
            def upd_unpack(d, v): return v.tolist() + [d, name]  # analysis:ignore
            def ins_unpack(d, v): return [d, name] + v.tolist()  # analysis:ignore

        else:
            def is_equal(v1, v2): return v1 == v2  # analysis:ignore
            def upd_unpack(d, v): return [v, d, name]  # analysis:ignore
            def ins_unpack(d, v): return [d, name, v]  # analysis:ignore

        ins, upd = [], []

        try:
            # --- get existing curve
            if crv_type == "HLOCV":
                curr_crv = self.get_hlocv(name, crv.front.date, crv.back.date)
            else:
                curr_crv = self.get_curve(name, crv.front.date, crv.back.date)
        except TsNotFound:
            ins = [ins_unpack(d, v) for d, v in crv]
        else:
            # --- loop over all knots and check which one needs updating
            #     and which one needs inserting
            for d, v in crv:
                try:
                    curr_val = curr_crv[d]
                    # --- knot exists already, update only if value has changed
                    #     NB: need to convert to a tuple to support comparison
                    #         of numpy arrays
                    if is_equal(curr_val, v):
                        upd.append(upd_unpack(d, v))
                except IndexError:
                    # --- knot not found, insert
                    ins.append(ins_unpack(d, v))

        if crv_type == "HLOCV":
            ins_query = "INSERT INTO HlocvCurves VALUES (%s,%s,%s,%s,%s,%s,%s)"
            upd_query = """UPDATE HlocvCurves
                           SET
                           High = %s,
                           Low = %s,
                           Open = %s,
                           Close = %s,
                           Volume = %s
                           WHERE Date = %s AND Name = %s"""
        else:
            ins_query = "INSERT INTO Curves VALUES (%s,%s,%s)"
            upd_query = """UPDATE Curves
                           SET
                           Value = %s
                           WHERE Date = %s AND Name = %s"""

        curs = self.conn.cursor()

        if len(ins):
            curs.executemany(ins_query, ins)
        if len(upd):
            curs.executemany(upd_query, upd)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..date import Date
from ..curve import Curve

import numpy as np
import unittest
import pickle


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- create a first curve by passing arrays of dates and values
        self.dates = [Date.parse(d)
                      for d in ["F08", "G08", "H08", "J08", "K08", "M08",
                                "N08", "Q08", "U08", "V08", "X08", "Z08"]]
        self.values = np.cumsum(np.ones(12))
        self.refcrv = Curve(self.dates, self.values)

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_algebra(self):
        values = self.refcrv.values
        self.assertEqual(2.0*self.refcrv - self.refcrv*1.0, self.refcrv)
        self.assertEqual(2.0*self.refcrv / 2.0, self.refcrv)
        self.assertEqual(sum(1.0 / values), sum((1.0 / self.refcrv).values))
        self.assertEqual(sum(values*values),
                         sum((self.refcrv*self.refcrv).values))

    def test_pickling(self):
        new_crv = pickle.loads(pickle.dumps(self.refcrv, -1))
        self.assertEqual(new_crv, self.refcrv)

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .ufo_base import UfoBase
from .objdb import ObjDbBase, ObjDbClient, ObjDbError

from .. import database as onyx_db

from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from psycopg2.extensions import ISOLATION_LEVEL_READ_COMMITTED
from psycopg2.extensions import ISOLATION_LEVEL_SERIALIZABLE

import getpass

__all__ = [
    "UseDatabase",
    "ObjDbTransaction",
    "ObjDbQuery",
    "AddObj",
    "GetObj",
    "UpdateObj",
    "DelObj",
    "ObjTypes",
    "ObjNamesByType",
    "ExistsInDatabase",
]


###############################################################################
class UseDatabase(object):
    """
    Context manager used to setup the objdb client used by the api.
    """
    # -------------------------------------------------------------------------
    def __init__(self, database, user=None):
        if isinstance(database, ObjDbBase):
            self.clt = database
        elif isinstance(database, str):
            self.clt = ObjDbClient(database, user or getpass.getuser())
        else:
            raise ValueError("UseDatabase only accepts instances of clients "
                             "derived from ObjDbBase or a valid database name")

    # -------------------------------------------------------------------------
    def __enter__(self):
        self.prev_clt = onyx_db.obj_clt
        self.prev_obj_instances = onyx_db.obj_instances
        onyx_db.obj_clt = self.clt
        onyx_db.obj_instances = dict()

    # -------------------------------------------------------------------------
    def __exit__(self, *args, **kwds):
        onyx_db.obj_clt.close()
        onyx_db.obj_instances.clear()
        onyx_db.obj_clt = self.prev_clt
        onyx_db.obj_instances = self.prev_obj_instances
        # --- returns False so that all execptions raised will be propagated
        return False


###############################################################################
class ObjDbTransaction(object):
    """
    Description:
        Transaction context manager. All statements executed within the context
        manager will be committed at the end or rolled back if an error occurs.
    Inputs:
        name   - the transaction unique name (per session)
        level  - choose between:
                    READ_COMMITTED (default)
                    SERIALIZABLE
        on_err - callback function executed in case the transaction is rolled
                 back.
    """
    # -------------------------------------------------------------------------
    def __init__(self, name, level="READ_COMMITTED", on_err=None):
        self.name = name
        if level == "READ_COMMITTED":
            self.level = ISOLATION_LEVEL_READ_COMMITTED
        elif level == "SERIALIZABLE":
            self.level = ISOLATION_LEVEL_SERIALIZABLE
        else:
            raise NameError("Unrecognized isolation level {0:s}".format(level))
        self.conn = onyx_db.obj_clt.conn
        # --- callback function called in case of exception
        self.on_err = on_err

    # -------------------------------------------------------------------------
    def __enter__(self):
        if onyx_db.obj_clt.transaction is None:
            if onyx_db.obj_clt.transaction != self.name:
                onyx_db.obj_clt.transaction = self.name
                self.conn.set_isolation_level(self.level)
                self.conn.tpc_begin(self.name)
            else:
                raise ObjDbError("Multiple nested transactions with same name")

    # -------------------------------------------------------------------------
    def __exit__(self, exc_type, exc_value, traceback):
        if onyx_db.obj_clt.transaction == self.name:
            if exc_type is None:
                self.conn.tpc_commit()
            else:
                self.conn.tpc_rollback()
                if self.on_err is not None:
                    self.on_err()
            # --- release transaction lock
            onyx_db.obj_clt.transaction = None
            # --- restore isolation level
            self.conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        # --- returns False so that all execptions raised will be propagated
        return False


# -----------------------------------------------------------------------------
def ObjDbQuery(query, parms=None, attr=""):
    """
    Description:
        Perform a db-query.
    Inputs:
        query - the SQL query as a valid string.
        parms - parameters, if any, that are passed to the query
        attr  - cursor attribute, such as fetchone or fetchall.
    Returns:
        The result of the query.
    """
    try:
        curs = onyx_db.obj_clt.conn.cursor()
    except AttributeError:
        raise ObjDbError("<{0:s}> does not support "
                         "queries".format(onyx_db.obj_clt.dbname))

    try:
        if parms is None:
            curs.execute(query)
        else:
            curs.execute(query, parms)
        return getattr(curs, attr, lambda: None)()
    finally:
        curs.close()


# -----------------------------------------------------------------------------
def AddObj(obj):
    """
    Description:
        Add a new UFO object to the database.
    Inputs:
        obj - can be either a UFO object or it's name. In the latter case,
              the object must have previously been created in memory.
    Returns:
        A reference to the added object.
    """
    # --- input validation
    if isinstance(obj, UfoBase):
        pass
    elif isinstance(obj, str):
        try:
            obj = onyx_db.obj_clt[obj]
        except KeyError:
            raise ObjDbError("No object to add with name {0:s}".format(obj))
    else:
        raise ObjDbError("Cannot persist objects that "
                         "don't inherit from UfoBase: {0!s}".format(type(obj)))

    return onyx_db.obj_clt.add(obj)


# -----------------------------------------------------------------------------
def GetObj(name, refresh=False):
    """
    Description:
        Retrieve an UFO object from the database. It first looks for the
        object in the database cache and finally in the database itself.
    Inputs:
        name    - the object's name.
        refresh - if True, ignore database cache.
    Returns:
        A reference to the looked-up object.
    """
    # --- input validation
    if not isinstance(name, str):
        raise ObjDbError("GetObj only retrives objects by name. "
                         "{0!s} was passed instead".format(type(name)))

    return onyx_db.obj_clt.get(name, refresh)


# -----------------------------------------------------------------------------
def UpdateObj(obj):
    """
    Description:
        Update the persisted version of an UFO object.
    Inputs:
        obj - can be either an UFO object or it's name. In the latter case, the
              object must have previously been modified by setting one of its
              attributes with SetVal. Therefore an instance of the object
              must exist in the graph.
    Returns:
        A reference to the updated object.
    """
    # --- input validation
    if isinstance(obj, UfoBase):
        pass
    elif isinstance(obj, str):
        # --- object passed by name: assume attributes have been set through
        #     the graph API
        try:
            obj = onyx_db.obj_clt[obj]
        except KeyError:
            raise ObjDbError("Nothing to update for {0:s}".format(obj))
    else:
        raise ObjDbError("UpdateObj must be fed either "
                         "an object derived from UfoBase or it's name")

    return onyx_db.obj_clt.update(obj)


# -----------------------------------------------------------------------------
def DelObj(obj):
    """
    Description:
        Deletes an UFO object from database. If available, the delete method
        of the object is called before deleting it.
    Inputs:
        obj - can be either an UFO object or it's name.
    Returns:
        None.
   """
    # --- make sure that the object exists in database
    if not isinstance(obj, UfoBase):
        obj = GetObj(obj)

    # --- we need a serializable transaction here to make sure that any action
    #     on dependent objects (see delete method of the object itself)
    #     is atomic
    with ObjDbTransaction("delete", level="SERIALIZABLE"):
        onyx_db.obj_clt.delete(obj)


# -----------------------------------------------------------------------------
def ExistsInDatabase(obj):
    """
    Description:
        Check if an object with the given name exists in database.
    Inputs:
        obj - can be either an UFO object or it's name.
    Returns:
        True if obj exists in database.
   """
    # --- input validation
    if isinstance(obj, UfoBase):
        obj = obj.Name
    elif not isinstance(obj, str):
        raise TypeError("ExistsInDatabase requires either a "
                        "string or an instance of a UfoBase class")

    return obj in onyx_db.obj_clt


# -----------------------------------------------------------------------------
def ObjTypes():
    """
    Description:
        Return a set with names of all object types stored in database.
    Returns:
        A set.
    """
    try:
        curs = onyx_db.obj_clt.conn.cursor()
    except AttributeError:
        raise ObjDbError("<{0:s}> does not "
                         "support queries".format(onyx_db.obj_clt.dbname))

    curs.execute("SELECT DISTINCT(ObjType) FROM ClassDefinitions;")
    return {row.objtype for row in curs.fetchall()}


# -----------------------------------------------------------------------------
def ObjNamesByType(obj_type=None):
    """
    Description:
        Return a list of all objects of the given type that are stored in
        database..
    Inputs:
        obj_type - the object type. If None, return all objects.
    Returns:
        A list.
   """
    try:
        curs = onyx_db.obj_clt.conn.cursor()
    except AttributeError:
        raise ObjDbError("<{0:s}> does not "
                         "support queries".format(onyx_db.obj_clt.dbname))

    if obj_type is None:
        curs.execute("SELECT Name FROM Objects;")
    else:
        curs.execute("SELECT Name "
                     "FROM Objects WHERE ObjType=%s;", (obj_type,))

    return [row.name for row in curs.fetchall()]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import FloatField, BoolField

from onyx.core.depgraph.graph_api import ValueType, CreateInMemory
from onyx.core.depgraph.graph_api import GetVal, GetNode
from onyx.core.depgraph.graph_scopes import EvalBlock, GraphScope

from onyx.core.utils.unittest import OnyxTestCase

import unittest

LEAF_1 = 12.3
LEAF_2 = 66.6
FLAG = True
X = 3.14


###############################################################################
class Dynamic(UfoBase):
    leaf1 = FloatField(default=LEAF_1)
    leaf2 = FloatField(default=LEAF_2)
    flag = BoolField(default=FLAG)

    # -------------------------------------------------------------------------
    @ValueType()
    def property_vt(self, graph):
        if graph(self, "flag"):
            return graph(self, "leaf1")
        else:
            return graph(self, "leaf2")

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def callable_vt(self, graph, x):
        if graph(self, "flag"):
            return graph(self, "leaf1") + x
        else:
            return graph(self, "leaf2") + x

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def prop_sub_graph_vt(self, graph, x=X):
        if graph(self, "flag"):
            return graph(self, "leaf1") + x
        else:
            return graph(self, "leaf2") + x


###############################################################################
class TestDynamic(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = CreateInMemory(Dynamic(Name="dynamic"))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def tearDown(self):
        super().tearDown()

    # -------------------------------------------------------------------------
    def test_leaves(self):
        self.assertEqual(GetVal(self.name, "leaf1"), LEAF_1)
        self.assertEqual(GetVal(self.name, "leaf2"), LEAF_2)
        self.assertEqual(GetVal(self.name, "flag"), FLAG)

    # -------------------------------------------------------------------------
    def test_property_vt(self):
        self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)

        with EvalBlock() as eb:
            eb.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name, "property_vt"), LEAF_2)

        self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)

        with EvalBlock() as eb:
            self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)
            eb.change_value(self.name, "leaf1", 2.71828)
            self.assertFalse(GetNode(self.name, "property_vt").valid)
            self.assertEqual(GetVal(self.name, "property_vt"), 2.71828)

        self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)

        with EvalBlock() as eb:
            self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)
            eb.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name, "property_vt"), LEAF_2)
            eb.change_value(self.name, "leaf1", 2.71828)
            self.assertTrue(GetNode(self.name, "property_vt").valid)

        self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)

        with EvalBlock() as eb:
            eb.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name, "property_vt"), LEAF_2)

        self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)

        with GraphScope() as scope:
            self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)
            scope.change_value(self.name, "leaf1", 2.71828)
            self.assertFalse(GetNode(self.name, "property_vt").valid)
            self.assertEqual(GetVal(self.name, "property_vt"), 2.71828)

        self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)

        with GraphScope() as scope:
            self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)
            scope.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name, "property_vt"), LEAF_2)
            scope.change_value(self.name, "leaf1", 2.71828)
            self.assertTrue(GetNode(self.name, "property_vt").valid)

        self.assertEqual(GetVal(self.name, "property_vt"), LEAF_1)

    # -------------------------------------------------------------------------
    def test_callable_vt(self):
        self.assertEqual(GetVal(self.name, "callable_vt", 1.0), LEAF_1 + 1.0)

        with EvalBlock() as eb:
            eb.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name,
                                    "callable_vt", 1.0), LEAF_2 + 1.0)

        self.assertEqual(GetVal(self.name, "callable_vt", 1.0), LEAF_1 + 1.0)

        with EvalBlock() as eb:
            self.assertEqual(GetVal(self.name,
                                    "callable_vt", 1.0), LEAF_1 + 1.0)
            eb.change_value(self.name, "leaf1", 2.71828)
            self.assertFalse(GetNode(self.name, "callable_vt", (1.0,)).valid)
            self.assertEqual(GetVal(self.name, "callable_vt", 1.0), 3.71828)

        self.assertEqual(GetVal(self.name, "callable_vt", 1.0), LEAF_1 + 1.0)

        with EvalBlock() as eb:
            self.assertEqual(GetVal(self.name,
                                    "callable_vt", 1.0), LEAF_1 + 1.0)
            eb.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name,
                                    "callable_vt", 1.0), LEAF_2 + 1.0)
            eb.change_value(self.name, "leaf1", 2.71828)
            self.assertTrue(GetNode(self.name, "callable_vt", (1.0,)).valid)

        self.assertEqual(GetVal(self.name, "callable_vt", 1.0), LEAF_1 + 1.0)

        with EvalBlock() as eb:
            eb.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name,
                                    "callable_vt", 1.0), LEAF_2 + 1.0)

        self.assertEqual(GetVal(self.name, "callable_vt", 1.0), LEAF_1 + 1.0)

        with GraphScope() as scope:
            self.assertEqual(GetVal(self.name,
                                    "callable_vt", 1.0), LEAF_1 + 1.0)
            scope.change_value(self.name, "leaf1", 2.71828)
            self.assertFalse(GetNode(self.name, "callable_vt", (1.0,)).valid)
            self.assertEqual(GetVal(self.name, "callable_vt", 1.0), 3.71828)

        self.assertEqual(GetVal(self.name, "callable_vt", 1.0), LEAF_1 + 1.0)

        with GraphScope() as scope:
            self.assertEqual(GetVal(self.name,
                                    "callable_vt", 1.0), LEAF_1 + 1.0)
            scope.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name,
                                    "callable_vt", 1.0), LEAF_2 + 1.0)
            scope.change_value(self.name, "leaf1", 2.71828)
            self.assertTrue(GetNode(self.name, "callable_vt", (1.0,)).valid)

        self.assertEqual(GetVal(self.name, "callable_vt", 1.0), LEAF_1 + 1.0)

    # -------------------------------------------------------------------------
    def test_property_sub_graph_vt(self):
        self.assertEqual(GetVal(self.name, "prop_sub_graph_vt"), LEAF_1 + X)
        self.assertEqual(GetVal(self.name,
                                "prop_sub_graph_vt", x=1.0), LEAF_1 + 1.0)

        with EvalBlock() as eb:
            self.assertEqual(GetVal(self.name,
                                    "prop_sub_graph_vt"), LEAF_1 + X)
            eb.change_value(self.name, "leaf1", 2.71828)
            self.assertFalse(GetNode(self.name, "prop_sub_graph_vt").valid)
            self.assertEqual(GetVal(self.name,
                                    "prop_sub_graph_vt"), 2.71828 + X)

        with EvalBlock() as eb:
            eb.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name,
                                    "prop_sub_graph_vt"), LEAF_2 + X)
            eb.change_value(self.name, "leaf1", 2.71828)
            self.assertTrue(GetNode(self.name, "prop_sub_graph_vt").valid)
            self.assertEqual(GetVal(self.name,
                                    "prop_sub_graph_vt"), LEAF_2 + X)

        self.assertEqual(GetVal(self.name, "prop_sub_graph_vt"), LEAF_1 + X)

        with GraphScope() as scope:
            self.assertEqual(GetVal(self.name,
                                    "prop_sub_graph_vt"), LEAF_1 + X)
            scope.change_value(self.name, "leaf1", 2.71828)
            self.assertFalse(GetNode(self.name, "prop_sub_graph_vt").valid)
            self.assertEqual(GetVal(self.name,
                                    "prop_sub_graph_vt"), 2.71828 + X)

        with GraphScope() as scope:
            scope.change_value(self.name, "flag", False)
            self.assertEqual(GetVal(self.name,
                                    "prop_sub_graph_vt"), LEAF_2 + X)
            scope.change_value(self.name, "leaf1", 2.71828)
            self.assertTrue(GetNode(self.name, "prop_sub_graph_vt").valid)
            self.assertEqual(GetVal(self.name,
                                    "prop_sub_graph_vt"), LEAF_2 + X)

        self.assertEqual(GetVal(self.name, "prop_sub_graph_vt"), LEAF_1 + X)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .objdb_api import ObjDbQuery, GetObj
from .ufo_base import get_base_classes, BaseField, MutableField, FieldError
from ..datatypes.date import Date
from ..datatypes.curve import Curve
from ..datatypes.gcurve import GCurve
from ..datatypes.hlocv import HlocvCurve
from ..datatypes.structure import Structure
from ..datatypes.table import Table
from ..datatypes.holiday_cal import HolidayCalendar

import pickle
import base64
import gzip
import numbers

__all__ = [
    "StringField",
    "SelectField",
    "IntField",
    "SelectIntField",
    "FloatField",
    "BoolField",
    "BinaryField",
    "DateField",
    "ListField",
    "DictField",
    "SetField",
    "CurveField",
    "GCurveField",
    "HlocvCurveField",
    "StructureField",
    "HolidayCalField",
    "ReferenceField",
]


# -----------------------------------------------------------------------------
# --- serialize to plain string using base64 encoding
def pickle_loads(val):
    return pickle.loads(gzip.decompress(base64.b64decode(val)))


def pickle_dumps(val):
    return base64.b64encode(gzip.compress(pickle.dumps(val))).decode("utf-8")


###############################################################################
class StringField(BaseField):
    def validate(self, val):
        if not isinstance(val, str):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class SelectField(StringField):
    def __init__(self, *args, **kwds):
        self.options = kwds.pop("options", None)
        super().__init__(*args, **kwds)
        if not all([isinstance(opt, str) for opt in self.options]):
            types = [type(opt) for opt in self.options]
            raise FieldError("One or more optins are not "
                             "of type str: {0!s}".format(types))

    def validate(self, val):
        super().validate(val)
        if val not in self.options:
            raise FieldError("Illegal value {0!s} for <{1!r}>. Valid options "
                             "are {2!s}".format(val, self, self.options))


###############################################################################
class IntField(BaseField):
    def __init__(self, *args, **kwds):
        self.positive = kwds.pop("positive", False)
        super().__init__(*args, **kwds)

    def validate(self, val):
        if not isinstance(val, int):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))
        if self.positive and val < 0:
            raise FieldError("Input for <{0!r}> is not "
                             "positive: {1:d}".format(self, val))


###############################################################################
class SelectIntField(IntField):
    def __init__(self, *args, **kwds):
        self.options = kwds.pop("options", None)
        super().__init__(*args, **kwds)
        if not all([isinstance(opt, int) for opt in self.options]):
            types = [type(opt) for opt in self.options]
            raise FieldError("One or more optins are not "
                             "of type int: {0!s}".format(types))

    def validate(self, val):
        super().validate(val)
        if val not in self.options:
            raise FieldError("Illegal value {0!s} for <{1!r}>. Valid options "
                             "are {2!s}".format(val, self, self.options))


###############################################################################
class FloatField(BaseField):
    def __init__(self, *args, **kwds):
        self.positive = kwds.pop("positive", False)
        super().__init__(*args, **kwds)

    def validate(self, val):
        if not isinstance(val, float):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))
        if self.positive and val < 0.0:
            raise FieldError("Input for <{0!r}> is "
                             "not positive: {1:f}".format(self, val))


###############################################################################
class BoolField(BaseField):
    def validate(self, val):
        if not isinstance(val, bool):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class BinaryField(BaseField):
    def from_json(self, val):
        return gzip.decompress(base64.b64decode(val))

    def to_json(self, val):
        return base64.b64encode(gzip.compress(val)).decode("utf-8")

    def validate(self, val):
        if not isinstance(val, bytes):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class DateField(BaseField):
    def from_json(self, val):
        return Date.parse(val)

    def to_json(self, val):
        return val.isoformat()

    def validate(self, val):
        if not isinstance(val, Date):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class ListField(MutableField):
    def validate(self, val):
        if not isinstance(val, list):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class DictField(MutableField):
    def validate(self, val):
        if not isinstance(val, dict):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class SetField(MutableField):
    def from_json(self, val):
        return val if val is None else set(val)

    def to_json(self, val):
        return val if val is None else list(val)

    def validate(self, val):
        if not isinstance(val, set):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class CurveField(MutableField):
    def from_json(self, val):
        if val is None:
            return val
        knts = [(Date.parse(r["date"]), r["value"]) for r in val]
        return Curve.create_raw([d for d, v in knts], [v for d, v in knts])

    def to_json(self, crv):
        if crv is None:
            return crv
        else:
            return [{"date": d.isoformat(), "value": v} for d, v in crv]

    def validate(self, val):
        if not isinstance(val, Curve):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class GCurveField(MutableField):
    def from_json(self, val):
        if val is None:
            return val
        knts = [(Date.parse(r["date"]), pickle_loads(r["value"])) for r in val]
        return GCurve.create_raw([d for d, v in knts], [v for d, v in knts])

    def to_json(self, crv):
        if crv is None:
            return crv
        else:
            return [{"date": d.isoformat(),
                     "value": pickle_dumps(v)} for d, v in crv]

    def validate(self, val):
        if not isinstance(val, GCurve):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class HlocvCurveField(MutableField):
    def from_json(self, val):
        if val is None:
            return val
        knts = [(Date.parse(row["date"]), row["value"]) for row in val]
        return HlocvCurve.create_raw([d for d, v in knts],
                                     [v for d, v in knts])

    def to_json(self, crv):
        if crv is None:
            return crv
        else:
            return [{"date": d.isoformat(),
                     "value": v.tolist()} for d, v in crv]

    def validate(self, val):
        if not isinstance(val, HlocvCurve):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class StructureField(MutableField):
    def from_json(self, val):
        return val if val is None else Structure(val)

    def to_json(self, val):
        return val if val is None else list(zip(val.keys(), val.values()))

    def validate(self, val):
        if not isinstance(val, Structure):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))
        # --- all keys should be strings
        if not all(isinstance(k, str) for k in val.keys()):
            FieldError("<{0!r}>, structure "
                       "keys need to be strings".format(self))
        # --- all values should be numbers
        if not all(isinstance(v, numbers.Number) for v in val.values()):
            FieldError("<{0!r}>, structure "
                       "values need to be numbers".format(self))


###############################################################################
class TableField(MutableField):
    def from_json(self, val):
        return val if val is None else pickle_loads(val)

    def to_json(self, val):
        return val if val is None else pickle_dumps(val)

    def validate(self, val):
        if not isinstance(val, Table):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class HolidayCalField(MutableField):
    def from_json(self, val):
        if val is None:
            return val
        else:
            return HolidayCalendar([Date.parse(d) for d in val])

    def to_json(self, val):
        return val if val is None else [d.isoformat() for d in val.holidays]

    def validate(self, val):
        if not isinstance(val, HolidayCalendar):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))


###############################################################################
class ReferenceField(StringField):
    def __init__(self, *args, **kwds):
        self.obj_type = kwds.pop("obj_type", None)
        super().__init__(*args, **kwds)

    def validate(self, val):
        # --- accept None as special value used to mark the reference as
        #     undefined
        if val is None:
            return
        if not isinstance(val, str):
            raise FieldError("Invalid input {0!s} "
                             "for <{1!r}>".format(type(val), self))
        if self.obj_type is None:
            res = ObjDbQuery("""SELECT EXISTS "
                                (SELECT 1 FROM Objects
                                 WHERE ObjName=%s) AS "exists";""",
                             parms=(val,), attr="fetchone")
            if res.exists:
                raise FieldError("Object {0:s} "
                                 "not found in database".format(val))
        else:
            bases = set(get_base_classes(GetObj(val).__class__))

            if self.obj_type not in bases:
                msg = ("Object {0:s} exists in database, "
                       "but it is not instance of a subclass of {1:s}")
                raise FieldError(msg.format(val, self.obj_type))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ...datatypes.table import Table
from ..table_fns import TableExtract

import unittest


# --- unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        self.table_array = [
            ("foo", "boo", "goo"),
            (3.0, "a", 5),
            (1.0, "b", 6),
            (2.0, "c", 7),
        ]
        self.table = Table(self.table_array)

    def tearDown(self):
        # perform clean-up actions, if any
        pass

    def test_table_extract(self):
        ref_table = Table([
            ("foo", "boo"),
            (3.0, "a"),
            (1.0, "b"),
            (2.0, "c"),
        ])
        self.assertEqual(TableExtract(self.table, ["foo", "boo"]), ref_table)

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import IntField
from onyx.core.database.objdb_api import AddObj, GetObj, DelObj
from onyx.core.depgraph.graph import GraphNode, SettableNode, PropertyNode
from onyx.core.depgraph.graph_api import ValueType, GetNode
from onyx.core.utils.unittest import OnyxTestCase
from onyx.core import database as onyx_db

import unittest

X = 1
Y = 2


###############################################################################
class UfoSimple(UfoBase):
    x = IntField(default=X)
    y = IntField(default=Y)

    # -------------------------------------------------------------------------
    @ValueType()
    def getsum(self, graph):
        return graph(self, "x") + graph(self, "y")

    # -------------------------------------------------------------------------
    def not_a_vt(self):
        pass


###############################################################################
class TestSimple(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = AddObj(UfoSimple(Name="simple"))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def tearDown(self):
        DelObj(self.name)
        super().tearDown()

    # -------------------------------------------------------------------------
    def test_references(self):
        # --- here we check that the object reference contained in each node
        #     is consistent with what is in the cache of the database client
        #     and survives refreshing such cache
        node = GetNode(self.name, "Name")
        ref_id = id(node.obj_ref())

        self.assertEqual(ref_id, id(self.instance))
        self.assertEqual(ref_id, id(GetObj(self.name)))

        del self.instance

        self.assertEqual(ref_id, id(GetObj(self.name)))
        self.assertEqual(ref_id, id(GetObj(self.name, True)))

        del onyx_db.obj_clt[self.name]
        with self.assertRaises(KeyError):
            onyx_db.obj_clt[self.name]

        self.assertNotEqual(ref_id, id(node.obj_ref()))  # obj_ref is None
        self.assertNotEqual(ref_id, id(GetObj(self.name, True)))

    # -------------------------------------------------------------------------
    def test_create_nodes(self):
        # --- create a node that points to "Name"
        node = GetNode(self.name, "Name")
        self.assertTrue(isinstance(node, SettableNode))
        self.assertEqual(node.get_value(), self.name)

        # --- create a node that points to "Version"
        node = GetNode(self.name, "Version")
        self.assertTrue(isinstance(node, SettableNode))
        self.assertEqual(node.get_value(), 0)

        # --- create a node that points to "StoredAttrs"
        node = GetNode(self.name, "StoredAttrs")
        self.assertTrue(isinstance(node, GraphNode))
        self.assertEqual(node.get_value(), self.instance.StoredAttrs)

        # --- create a node that points to a method decorated by ValueType
        node = GetNode(self.name, "getsum")
        self.assertTrue(isinstance(node, PropertyNode))
        self.assertEqual(node.get_value(), self.instance.getsum)

        # --- create a node that points to a method that is not a ValueType
        with self.assertRaises(AttributeError):
            GetNode(self.name, "not_a_vt")

        # --- create a node that points to a non-existing method
        with self.assertRaises(AttributeError):
            GetNode(self.name, "not_there")

    # -------------------------------------------------------------------------
    def test_clone(self):
        # --- create a node that points to "StoredAttrs" and make a clone
        node1 = GetNode(self.name, "StoredAttrs")
        node2 = node1.clone()

        # --- test that the two nodes are different objects but return the same
        #     value and share the same reference to the underlying object
        self.assertNotEqual(node1, node2)
        self.assertEqual(node1.get_value(), node2.get_value())
        self.assertEqual(node1.obj_ref, node2.obj_ref)
        self.assertIs(node1.obj_ref, node2.obj_ref)
        self.assertIs(node1.obj_ref(), self.instance)
        self.assertIs(node2.obj_ref(), self.instance)

        # --- create a node that points to "sum" and make a clone
        node1 = GetNode(self.name, "getsum")
        node2 = node1.clone()

        # --- test that the two nodes are different objects but return the same
        #     value and share the same reference to the underlying object
        self.assertNotEqual(node1, node2)
        self.assertEqual(node1.get_value(), node2.get_value())
        self.assertEqual(node1.obj_ref, node2.obj_ref)
        self.assertIs(node1.obj_ref, node2.obj_ref)
        self.assertIs(node1.obj_ref(), self.instance)
        self.assertIs(node2.obj_ref(), self.instance)

        # --- test that invalidation only affects a specific node
        x = GetNode(self.name, "x")
        x.value = 6

        node1.invalidate()
        self.assertFalse(node1.valid)
        self.assertTrue(node2.valid)
        self.assertEqual(node1.get_value(), 6 + Y)
        self.assertEqual(node2.get_value(), X + Y)
        node2.invalidate()
        self.assertFalse(node2.valid)
        self.assertEqual(node2.get_value(), 6 + Y)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .graph import DependencyGraph, GraphError, PropertyNode, SettableNode
from .graph import create_node

from .. import depgraph as onyx_dg

import collections

__all__ = ["EvalBlock", "GraphScope"]


###############################################################################
class EvalBlock(object):
    """
    Description:
        Context manager used to manage lifetime of one or more one-off changes.
    Usage:
        Typical use is as follows:

        with EvalBlock() as eb:
            ...
            eb.change_value("abc", "xyz", 123)
            ...
    """
    # -------------------------------------------------------------------------
    def __init__(self):
        self.__changes = []
        self.__changed_nodes = {}

    # -------------------------------------------------------------------------
    def __enter__(self):
        # --- return a reference to itself (to be used by change_value)
        return self

    # -------------------------------------------------------------------------
    def __exit__(self, *args, **kwds):
        for node_id in reversed(self.__changes):
            old_node = self.__changed_nodes[node_id]

            # --- invalidate all parents, but not the node itself
            for parent in onyx_dg.graph[node_id].parents | old_node.parents:
                onyx_dg.graph[parent].invalidate()

            # --- reconstruct connection with children
            for child in old_node.children:
                onyx_dg.graph[child].parents.add(node_id)

            # --- store pre-change node into the graph
            onyx_dg.graph[node_id] = old_node

        self.__changes = []
        self.__changed_nodes = {}

        # --- returns False so that all execptions raised will be propagated
        return False

    # -------------------------------------------------------------------------
    def change_value(self, obj, VT, value):
        """
        Description:
            Change the in-memory value of a ValueType within an EvalBlock.
        Inputs:
            obj   - instance (or name) of an object derived from UfoBase
            VT    - name of the target ValueType
            value - the new value for the VT
        Returns:
            None.
        """
        name = obj if isinstance(obj, str) else obj.Name
        node_id = (name, VT, ())
        new_node = False

        try:
            node = onyx_dg.graph[node_id]
        except KeyError:
            new_node = True
            onyx_dg.graph[node_id] = node = create_node(node_id)

        if not isinstance(node, (PropertyNode, SettableNode)):
            raise GraphError("Unsupported node type: "
                             "{0:s}".format(node.__class__.__name__))

        # --- copy pre-change node and store it in __changed_nodes. This is
        #     only required the first time a VT is changed.
        if node_id not in self.__changed_nodes:
            self.__changes.append(node_id)
            self.__changed_nodes[node_id] = node.clone()

        if not new_node:
            # --- discard this node from the set of parents of all its children
            for child in node.children:
                onyx_dg.graph[child].parents.discard(node_id)

            # --- remove all its children
            node.children.clear()

            # --- invalidate all its parents
            for parent in node.parents:
                onyx_dg.graph[parent].invalidate()

        # --- set the node value and set its state to valid
        node.value = value
        node.valid = True


###############################################################################
class dict_with_fallback(collections.UserDict):
    # -------------------------------------------------------------------------
    def __init__(self, fallback, *args, **kwds):
        super().__init__(*args, **kwds)
        self.fallback = fallback

    # -------------------------------------------------------------------------
    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            return self.fallback[item]


###############################################################################
class GraphScope(DependencyGraph):
    """
    Description:
        Context manager used to manage the lifetime of in-memory changes to
        ValueTypes. GraphScope can be used to create persistent scenarions that
        are then re-used multiple times.
    Usage:
        Typical use is as follows:

        scope = GraphScope()
        scope.change_value("abc", xyz", 123)
        with scope:
            ...
            scope.change_value("abc", "xxx", 666)
            ...

        with scope:
            ...
            ...
    """
    # -------------------------------------------------------------------------
    def __init__(self):
        super().__init__()
        # --- the GraphScope keeps track of references to the underlying graph
        #     at the time when a new GraphScope instance is created
        self.udl_graph = onyx_dg.graph
        self.vts_by_obj = dict_with_fallback(onyx_dg.graph.vts_by_obj)

        # --- used to determine if the GraphScope is being used as a context
        #     manager
        self.__active = False
        # --- when GraphScope is being used as a context manager it becomes
        #     the global graph instance and a reference to the previous graph
        #     instance is stored here.
        self.__last_graph = None

    # -------------------------------------------------------------------------
    def __enter__(self):
        self.__active = True
        self.__last_graph = onyx_dg.graph
        onyx_dg.graph = self
        # --- return a reference to itself (to be used for disposable scopes)
        return self

    # -------------------------------------------------------------------------
    def __exit__(self, *args, **kwds):
        onyx_dg.graph = self.__last_graph
        self.__active = False
        self.__last_graph = None
        # --- returns False so that all execptions raised will be propagated
        return False

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = self.__new__(self.__class__)
        clone.data = self.data.copy()
        clone.udl_graph = self.udl_graph
        clone.vts_by_obj = self.vts_by_obj.copy()
        clone.__active = self.__active
        clone.__last_graph = self.__last_graph
        memo[id(self)] = clone
        return clone

    # -------------------------------------------------------------------------
    def __getitem__(self, node_id):
        try:
            return super().__getitem__(node_id)
        except KeyError:
            return self.udl_graph[node_id]

    # -------------------------------------------------------------------------
    def __copy_and_invalidate(self, node_id):
        try:
            node = super().__getitem__(node_id)
            node.valid = False
        except KeyError:
            try:
                node = self.udl_graph[node_id].clone()
                node.valid = False
            except KeyError:
                node = create_node(node_id)

            self[node_id] = node

        for parent in node.parents:
            self.__copy_and_invalidate(parent)

        return node

    # -------------------------------------------------------------------------
    def change_value(self, obj, VT, value):
        """
        Description:
            Change the value of a ValueType within a GraphScope.
            NB: only StoredAttrs and VTs defined as Property can be changed.
        Inputs:
            obj   - instance (or name) of an object derived from UfoBase
            VT    - name of the target ValueType
            value - the new value for the VT
        Returns:
            None.
        """
        if not self.__active:
            # --- switch graph to graph scope
            self.__last_graph = onyx_dg.graph
            onyx_dg.graph = self

        name = obj if isinstance(obj, str) else obj.Name
        node_id = (name, VT, ())

        # --- create a local copy (local to the GraphScope) of the sub-graph
        #     that depends on the changed node as well as the immediate
        #     children of the copied nodes.
        #     NB: all nodes are created anew, they are invalid and sub-graph
        #         topology is undefined.
        node = self.__copy_and_invalidate(node_id)

        if not isinstance(node, (PropertyNode, SettableNode)):
            raise GraphError("Unsupported node type: "
                             "{0:s}".format(node.__class__.__name__))

        # --- set the node value and set its state to valid
        node.value = value
        node.valid = True

        if not self.__active:
            # --- switch graph back to the fallback
            onyx_dg.graph = self.__last_graph
            self.__last_graph = None
